const double VERSION=0.7;

#pragma comment(lib,"lua5.1.lib")
#pragma comment(lib,"libpng.lib")
#pragma comment(lib,"vfw32.lib")
#include "lua.hpp"
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <share.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <vfw.h>
#include "png.h"
#include "pngconf.h"
#include <math.h>
#include <map>
#include <string>
#include <direct.h>
using namespace std;
const int DLLMAX=1024;
class DLLManager {
public:
	HINSTANCE inst[DLLMAX];
	string name[DLLMAX];
	DLLManager() {
		for (int i=0;i<DLLMAX;i++) name[i].clear();
	}
	~DLLManager() {
		for (int i=0;i<DLLMAX;i++) {
			if (!name[i].empty()) {
				FreeLibrary(inst[i]);
			}
		}
	}
	void *Get (const char *dname,const char *procname)  {
		int i;
		char dllname[256];
		for (i=0;dname[i]!='\0';i++) {
			dllname[i]=tolower(dname[i]);
		}
		for (i=0;i<DLLMAX;i++) {
			if (name[i].empty()) {
				break;
			}
			if (strcmp(name[i].c_str(),dllname)==0) {
				return GetProcAddress (inst[i],procname);
			}
		}
		if (i==DLLMAX) return NULL;
		inst[i]=LoadLibrary(dllname);
		name[i]=dllname;
		return GetProcAddress (inst[i],procname);
	}
};
DLLManager dm;

enum BlendMode {
	BM_COPY,
	BM_ALPHA,
	BM_ADD,
};

BlendMode blendmode;

int filter;

class DIB {
	//bpp32bitŒAΉtH[}bgpnĝ
public:
	BYTE *lpbits;
	BITMAPINFO *pbmi;
	int width,height;
	DIB() {pbmi=NULL;lpbits=NULL;}
	~DIB() {Delete();}
	void Delete (void) {
		if (pbmi!=NULL) {
			free(pbmi);
			pbmi=NULL;
		}
	}
	void Clear (DWORD argb) {
		DWORD *ptr=(DWORD *)lpbits;
		int sz=width*height;
		for (int i=0;i<sz;i++) {
			ptr[i]=argb;
		}
	}
	void Fill (int lx,int ly,int wd,int ht,DWORD argb) {
		if (lx<0) {wd+=lx;lx=0;}
		if (ly<0) {ht+=ly;ly=0;}
		if (lx+wd>width) {wd=width-lx;}
		if (ly+ht>height) {ht=height-ly;}
		DWORD *ptr=(DWORD *)(lpbits+lx*4+ly*width*4);
		int pitch=width;
		for (int j=0;j<ht;j++) {
			for (int i=0;i<wd;i++) {
				ptr[i]=argb;
			}
			ptr=ptr+pitch;
		}
	}
	bool Load (const char *filename) {
		char buf[16];
		int l=strlen(filename);
		while(l>0 && filename[l]!='.') l--;
		strcpy_s(buf,16,filename+l+1);
		for (int i=0;buf[i]!='\0';i++) buf[i]=tolower(buf[i]);
		if (strcmp(buf,"png")==0) {
			return LoadPNG(filename);
		} else if (strcmp(buf,"bmp")==0) {
			return LoadBMP(filename);
		} else if (strcmp(buf,"jpeg")==0 || strcmp(buf,"jpg")==0) {
			return LoadJPEG(filename);
		} else {
			return false;
		}
	}
	bool LoadPNG (const char *filename);
	bool LoadBMP (const char *filename);
	bool LoadJPEG (const char *filename);
	bool Save (const char *filename);
	bool Duplicate (DIB *src);
	bool Create (int w,int h);
	void Render (int x,int y,int w,int h,DIB *src,int x2,int y2,int srcalpha,BlendMode md);
	void Render2_0 (
		int dcx,int dcy,
		DIB *src,int slx,int sly,int sw,int sh,
		double xs,double ys,double rot,
		int srcalpha,BlendMode md
	);
	void Render2_1 (
		int dcx,int dcy,
		DIB *src,int slx,int sly,int sw,int sh,
		double xs,double ys,double rot,
		int srcalpha,BlendMode md
	);
	void Render2_2 (
		int dcx,int dcy,
		DIB *src,int slx,int sly,int sw,int sh,
		double xs,double ys,double rot,
		int srcalpha,BlendMode md
	);
	void Render2_3 (
		int dcx,int dcy,
		DIB *src,int slx,int sly,int sw,int sh,
		double xs,double ys,double rot,
		int srcalpha,BlendMode md
	);

};

struct matrix {
	double _00,_10;
	double _01,_11;
};
inline void mulmatrix (double *ox,double *oy,double ix,double iy,matrix *m) {
	*ox=ix*m->_00+iy*m->_01;
	*oy=ix*m->_10+iy*m->_11;
}

int main (int argc,char **argv);
void LuaRegister (void);
lua_State *L;

inline double L1 (double x0,double p0,double x1,double p1,double x) {
	if (p0==p1) return p0;
	if (x0==x1) {
		luaL_error(L,"`ԂɎs܂BsȒlw肵Ă܂B");
		return 0.0;
	}
	return (x-x0)*p1/(x1-x0)+(x-x1)*p0/(x0-x1);
}

bool outformat;
class AVI {
public:
    LPBITMAPINFOHEADER pbmih;
    PAVIFILE pavi;
    PAVISTREAM pstm;
    PGETFRAME pfrm;
	AVI() {
		pavi=NULL;
		pstm=NULL;
		pfrm=NULL;
		pbmih=NULL;
	}
	~AVI() {Delete();}
	void Delete(void) {
		if (pfrm) {
			AVIStreamGetFrameClose(pfrm);
			pfrm=NULL;
		}
		if (pstm) {
			AVIStreamClose(pstm);
			pstm=NULL;
		}
		if (pavi) {
			AVIFileRelease(pavi);
			pavi=NULL;
		}
	}
	bool Load (const char *filename) {
		Delete();
		if (AVIFileOpen(&pavi,filename,OF_READ | OF_SHARE_DENY_NONE,NULL)!=0) {
			luaL_error(L,"AVIt@C\"%s\"ǂݍ߂܂łB",filename);
			return false;
		}
		if (AVIFileGetStream(pavi,&pstm,0,0)!=0) {
			luaL_error(L,"AVIt@C\"%s\"̃Xg[J܂łB",filename);
			return false;
		}
		if ((pfrm=AVIStreamGetFrameOpen(pstm,0))==NULL) {
			luaL_error(L,"AVIt@C\"%s\"̃Xg[t[擾ł܂B",filename);
			return false;
		}
		printf ("AVIt@C\"%s\"͗pɊJ܂B\n",filename);
		return true;
	}
	bool GetFrame (DIB *dib,int n) {
		//ݐrbg}bv͏32bit
		if (!pfrm) {
			luaL_error(L,"AVIXg[t[擾ł܂B");
			return false;
		}
		pbmih=(LPBITMAPINFOHEADER)AVIStreamGetFrame(pfrm,n);
		if (pbmih==NULL) {
			luaL_error(L,"t[̎擾Ɏs܂B");
			return false;
		}

		dib->Create(pbmih->biWidth,pbmih->biHeight);
		if (pbmih->biBitCount==32) {
			//㉺𔽓]B
			int ofs=(pbmih->biWidth)*4;
			unsigned char *sbits=(unsigned char *)(pbmih)+sizeof(BITMAPINFOHEADER);
			int dp=0;
			int sp=ofs*(pbmih->biHeight-1);
			for (int i=0;i<pbmih->biHeight;i++) {
				memcpy(dib->lpbits+dp,sbits+sp,ofs);
				sp-=ofs;
				dp+=ofs;
			}
		} else if (pbmih->biBitCount==24) {
			//2432ϊA㉺𔽓]B
			int dofs=pbmih->biWidth*4;
			int sofs=(((pbmih->biWidth*3)+3)/4)*4;
			unsigned char *sbits=(unsigned char *)(pbmih)+sizeof(BITMAPINFOHEADER);
			int dp=0;
			int sp=sofs*(pbmih->biHeight-1);
			for (int i=0;i<pbmih->biHeight;i++) {
				for (int j=0;j<pbmih->biWidth;j++) {
					dib->lpbits[dp+j*4]=sbits[sp+j*3];
					dib->lpbits[dp+j*4+1]=sbits[sp+j*3+1];
					dib->lpbits[dp+j*4+2]=sbits[sp+j*3+2];
					dib->lpbits[dp+j*4+3]=0xFF;
				}
				sp-=sofs;
				dp+=dofs;
			}
		} else {
			luaL_error (L,"AVIt@C24bit32bit̉摜`ΉĂ܂B");
			return false;
		}
		return true;
	}
};


typedef map<string,DIB *> DIBmap;
DIBmap dibmap;

typedef map<string,AVI *> AVImap;
AVImap avimap;


AVISTREAMINFO si;
PAVIFILE pavi;
PAVISTREAM pstm;
int sizeimage;
int framenum;
int fps;
int width;
int height;
bool startflag=false;
void WriteStart (const char *filename,LPBITMAPINFOHEADER lpbi,int fps,int fnum);
void Write (int frame,unsigned char *lpbits);
void WriteEnd (void);

bool DIB::LoadPNG (const char *filename) {
	Delete();
	png_structp png_ptr;
	png_infop info_ptr;
	png_uint_32 wd,ht;
	int bit_depth,color_type,interlace_type;
	FILE *fp;
	if (fopen_s(&fp,filename,"rb")!=0) return false;
	png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
	info_ptr=png_create_info_struct(png_ptr);
	png_init_io(png_ptr,fp);
	png_read_info(png_ptr,info_ptr);
	png_get_IHDR(png_ptr,info_ptr,&wd,&ht,&bit_depth,&color_type,&interlace_type,NULL,NULL);
	png_set_bgr(png_ptr);
	if (color_type==PNG_COLOR_TYPE_RGBA) {
		//̂܂܂ł
	} else if (color_type==PNG_COLOR_TYPE_RGB) {
		png_set_add_alpha(png_ptr,0xFF, PNG_FILLER_AFTER);
	} else {
		if (((color_type == PNG_COLOR_TYPE_GRAY) || (color_type == PNG_COLOR_TYPE_GRAY_ALPHA))) {
			if (bit_depth<8) {png_set_gray_1_2_4_to_8(png_ptr);}
			png_set_gray_to_rgb(png_ptr);
			if (color_type == PNG_COLOR_TYPE_GRAY) {
				png_set_add_alpha(png_ptr,0xFF, PNG_FILLER_AFTER);
			}
		}
		if (color_type == PNG_COLOR_TYPE_PALETTE) {
			png_set_palette_to_rgb(png_ptr);
			png_set_add_alpha(png_ptr,0xFF, PNG_FILLER_AFTER);
		}
		if (png_get_valid(png_ptr, info_ptr,PNG_INFO_tRNS)) {
			png_set_tRNS_to_alpha(png_ptr);
		}
	}
	width=wd;height=ht;
	pbmi=(BITMAPINFO *)malloc(width*height*4+sizeof(BITMAPINFO));
	pbmi->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biSizeImage=width*height*4;
	pbmi->bmiHeader.biBitCount=32;
	pbmi->bmiHeader.biClrImportant=0;
	pbmi->bmiHeader.biClrUsed=0;
	pbmi->bmiHeader.biCompression=BI_RGB;
	pbmi->bmiHeader.biHeight=-height;
	pbmi->bmiHeader.biWidth=width;
	pbmi->bmiHeader.biPlanes=1;
	pbmi->bmiHeader.biXPelsPerMeter=0;
	pbmi->bmiHeader.biYPelsPerMeter=0;
	lpbits=(BYTE *)(&(pbmi->bmiColors[0]));
	int pitch=width*4;
	png_bytep row_pointer=(png_bytep)(lpbits);
	for (int i=0;i<height;i++) {
		png_read_row(png_ptr, row_pointer, NULL);
		row_pointer+=pitch;
	}
	png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
	fclose(fp);
	return true;
}
bool DIB::Create (int w,int h) {
	Delete();
	width=w;height=h;
	pbmi=(BITMAPINFO *)malloc(width*height*4+sizeof(BITMAPINFO));
	pbmi->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biSizeImage=width*height*4;
	pbmi->bmiHeader.biBitCount=32;
	pbmi->bmiHeader.biClrImportant=0;
	pbmi->bmiHeader.biClrUsed=0;
	pbmi->bmiHeader.biCompression=BI_RGB;
	pbmi->bmiHeader.biHeight=-height;
	pbmi->bmiHeader.biWidth=width;
	pbmi->bmiHeader.biPlanes=1;
	pbmi->bmiHeader.biXPelsPerMeter=0;
	pbmi->bmiHeader.biYPelsPerMeter=0;
	lpbits=(BYTE *)(&(pbmi->bmiColors[0]));
	return true;
}

bool DIB::LoadJPEG (const char *filename) {
	printf ("o[WłJPEGɑΉĂ܂B\n");
	return false;
}


bool DIB::LoadBMP (const char *filename) {
	Delete();
	bool rev=false;
	FILE *fp;
	if (fopen_s(&fp,filename,"rb")!=0) return false;
	fseek(fp,sizeof(BITMAPFILEHEADER),SEEK_SET);
	BITMAPINFOHEADER bih;
	fread(&bih,sizeof(BITMAPINFOHEADER),1,fp);
	if (bih.biClrUsed>0) {
		fseek(fp,bih.biClrUsed*sizeof(RGBQUAD),SEEK_CUR);
	}
	width=bih.biWidth;
	height=bih.biHeight;
	if (bih.biBitCount<24) {
		printf("[hG[F24bit32bitBMPt@CɂΉĂ܂B\n");
		return false;
	}
	if (height<0) {
		height=-height;
		rev=true;
	}
	pbmi=(BITMAPINFO *)malloc(width*height*4+sizeof(BITMAPINFO));
	pbmi->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biSizeImage=width*height*4;
	pbmi->bmiHeader.biBitCount=32;
	pbmi->bmiHeader.biClrImportant=0;
	pbmi->bmiHeader.biClrUsed=0;
	pbmi->bmiHeader.biCompression=BI_RGB;
	pbmi->bmiHeader.biHeight=-height;
	pbmi->bmiHeader.biWidth=width;
	pbmi->bmiHeader.biPlanes=1;
	pbmi->bmiHeader.biXPelsPerMeter=0;
	pbmi->bmiHeader.biYPelsPerMeter=0;
	lpbits=(BYTE *)(&(pbmi->bmiColors[0]));
	if (rev) {
		if (bih.biBitCount==32) {
			int ofs=width*4;
			unsigned char *line=lpbits;
			for (int j=0;j<height;j++) {
				for (int i=0;i<width;i++) {
					line[i*4+3]=255;
					fgetc(fp);//ǂݎ̂
					fread(line+i*4,3,1,fp);
				}
				line+=ofs;
			}
		} else {
			int ofs=width*4;
			unsigned char *line=lpbits;
			for (int j=0;j<height;j++) {
				for (int i=0;i<width;i++) {
					fread(line+i*4,3,1,fp);
					line[i*4+3]=255;
				}
				line+=ofs;
			}
		}
	} else {
		if (bih.biBitCount==32) {
			int ofs=width*4;
			unsigned char *line=lpbits+(height-1)*ofs;
			for (int j=0;j<height;j++) {
				for (int i=0;i<width;i++) {
					line[i*4+3]=255;
					fgetc(fp);//ǂݎ̂
					fread(line+i*4,3,1,fp);
				}
				line-=ofs;
			}
		} else {
			int ofs=width*4;
			unsigned char *line=lpbits+(height-1)*ofs;
			for (int j=0;j<height;j++) {
				for (int i=0;i<width;i++) {
					fread(line+i*4,3,1,fp);
					line[i*4+3]=255;
				}
				line-=ofs;
			}
		}
	}
	fclose(fp);
	return true;
}


void DIB::Render (int x,int y,int w,int h,DIB *src,int x2,int y2,int srcalpha,BlendMode md) {
	if (srcalpha<=0) return;
	if (srcalpha>255) srcalpha=255;
	if (pbmi==NULL || src->pbmi==NULL) return;
	if (x>width || y>height) return;
	if (x2>src->width || y2>src->height) return;
	if (x<0) {w+=x;x2-=x;x=0;}
	if (y<0) {h+=y;y2-=y;y=0;}
	if (x+w>=width) {w=width-x;}
	if (y+h>=height) {h=height-y;}
	if (x2+w>=src->width) {w=src->width-x2;}
	if (y2+h>=src->height) {h=src->height-y2;}
	if (w<0 || h<0) return ;
	int pitch=width*4;
	int spitch=src->width*4;
	int destxs=x*4+y*pitch;
	int srcxs=x2*4+y2*spitch;
	unsigned char *d=lpbits+destxs;
	unsigned char *s=src->lpbits+srcxs;
	int a,c;
	int wp=w*4;
	if (md==BM_ALPHA) {
		for (int j=0;j<h;j++) {
			for (int i=0;i<wp;i+=4) {
				a=s[i+3]*srcalpha/255;
				d[i]=((int)s[i]-(int)d[i])*a/255+(int)d[i];
				d[i+1]=((int)s[i+1]-(int)d[i+1])*a/255+(int)d[i+1];
				d[i+2]=((int)s[i+2]-(int)d[i+2])*a/255+(int)d[i+2];
				d[i+3]=255;
			}
			d+=pitch;
			s+=spitch;
		}
	} else if (md==BM_ADD) {
		for (int j=0;j<h;j++) {
			for (int i=0;i<wp;i+=4) {
				a=s[i+3]*srcalpha/255;
				c=d[i]+s[i]*a/255;
				d[i]=(c>255)?255:c;
				c=d[i+1]+s[i+1]*a/255;
				d[i+1]=(c>255)?255:c;
				c=d[i+2]+s[i+2]*a/255;
				d[i+2]=(c>255)?255:c;
				d[i+3]=255;
			}
			d+=pitch;
			s+=spitch;
		}
	} else {
		//͕pɂɎĝŋCgč
		DWORD *dd=(DWORD *)d;
		DWORD *ss=(DWORD *)s;
		for (int j=0;j<h;j++) {
			for (int i=0;i<w;i++) {
				dd[i]=ss[i];
			}
			dd+=width;
			ss+=src->width;
		}
	}
}

bool DIB::Save (const char *filename) {
	FILE *fp;
	png_structp png_ptr;
	png_infop info_ptr;
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!png_ptr) return false;
	info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr){
        png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
        return false;
    }
	if (fopen_s(&fp,filename, "wb")!=0) {
		return false;
	}
	png_init_io(png_ptr, fp);
	png_set_IHDR(png_ptr,info_ptr,width,height,8,PNG_COLOR_TYPE_RGB_ALPHA,PNG_INTERLACE_NONE,PNG_COMPRESSION_TYPE_DEFAULT,PNG_FILTER_TYPE_BASE);
	png_color_8 sig_bit;
	sig_bit.red=8;
	sig_bit.green=8;
	sig_bit.blue=8;
	sig_bit.alpha=8;
	png_set_sBIT(png_ptr,info_ptr,&sig_bit);
	png_write_info(png_ptr,info_ptr);
	png_set_bgr(png_ptr);
	png_bytep row_pointer=(png_bytep)lpbits;
	int pitch=width*4;
	for (int i=0;i<height;i++) {
		png_write_row(png_ptr,row_pointer);
		row_pointer+=pitch;
	}
    png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr,&info_ptr);
	fclose(fp);
	return true;
}

bool DIB::Duplicate (DIB *src) {
	Delete();
	if (!Create(src->width,src->height)) return false;
	memcpy(lpbits,src->lpbits,width*height*4);
	return true;
}


//LuaĂяo֐
int Out(lua_State *L);
int OutFormat(lua_State *L);
int Frame(lua_State *L);
int Log(lua_State *L);
int TLoad(lua_State *L);
int TCreate(lua_State *L);
int TSave(lua_State *L);
int TDelete(lua_State *L);
int TClear(lua_State *L);
int TFill(lua_State *L);
int TDup(lua_State *L);
int TPut(lua_State *L);
int TPut2(lua_State *L);
int TInfo(lua_State *L);
int TBlend(lua_State *L);
int VOpen(lua_State *L);
int VGet(lua_State *L);
int VClose(lua_State *L);
int VInfo(lua_State *L);
int OutInfo(lua_State *L);
int DLLBind(lua_State *L);
int GetVersion(lua_State *L);
int Filter(lua_State *L);
int Lag(lua_State *L);
int Spl(lua_State *L);
int L1_l(lua_State *L);
//܂

//ODLLɌJ֐

extern "C" {
//Luå֐
	__declspec(dllexport) int fes_error(void *L,const char *errorstr);
	__declspec(dllexport) void fes_settop(void *L,int index);
	__declspec(dllexport) int fes_isnumber(void *L,int index);
	__declspec(dllexport) int fes_istable(void *L,int index);
	__declspec(dllexport) int fes_isstring(void *L,int index);
	__declspec(dllexport) int fes_isboolean(void *L,int index);
	__declspec(dllexport) int fes_isnil(void *L,int index);
	__declspec(dllexport) double fes_tonumber(void *L,int index);
	__declspec(dllexport) int fes_tointeger(void *L,int index);
	__declspec(dllexport) const char *fes_tostring(void *L,int index);
	__declspec(dllexport) int fes_toboolean(void *L,int index);
	__declspec(dllexport) void fes_pushnumber(void *L,double n);
	__declspec(dllexport) void fes_pushinteger(void *L,int n);
	__declspec(dllexport) void fes_pushstring(void *L,const char *str);
	__declspec(dllexport) void fes_pushboolean(void *L,int b);
	__declspec(dllexport) void fes_gettable(void *L,int index);
	__declspec(dllexport) void fes_pop(void *L,int index);

//fes̊֐
	__declspec(dllexport) bool fes_GetTexture(BITMAPINFOHEADER *bmih,unsigned char **bits,const char *texname);
}

int fes_error(void *L,const char *errorstr){return luaL_error((lua_State *)L,"%s",errorstr);}
void fes_settop(void *L,int index){lua_settop((lua_State *)L,index);}
int fes_isnumber(void *L,int index){return lua_isnumber((lua_State *)L,index);}
int fes_istable(void *L,int index){return lua_istable((lua_State *)L,index);}
int fes_isstring(void *L,int index){return lua_isstring((lua_State *)L,index);}
int fes_isboolean(void *L,int index){return lua_isboolean((lua_State *)L,index);}
int fes_isnil(void *L,int index){return lua_isnil((lua_State *)L,index);}
double fes_tonumber(void *L,int index){return lua_tonumber((lua_State *)L,index);}
int fes_tointeger(void *L,int index){return lua_tointeger((lua_State *)L,index);}
const char *fes_tostring(void *L,int index){return lua_tostring((lua_State *)L,index);}
int fes_toboolean(void *L,int index){return lua_toboolean((lua_State *)L,index);}
void fes_pushnumber(void *L,double n){lua_pushnumber((lua_State *)L,n);}
void fes_pushinteger(void *L,int n){lua_pushinteger((lua_State *)L,n);}
void fes_pushstring(void *L,const char *str){lua_pushstring((lua_State *)L,str);}
void fes_pushboolean(void *L,int b){lua_pushboolean((lua_State *)L,b);}
void fes_gettable(void *L,int index){lua_gettable((lua_State *)L,index);}
void fes_pop(void *L,int index){lua_pop((lua_State *)L,index);}
bool fes_GetTexture (BITMAPINFOHEADER *bmih,unsigned char **bits,const char *texname) {
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		return false;
	}
	*bmih=itr->second->pbmi->bmiHeader;
	*bits=itr->second->lpbits;
	return true;
}
void LuaRegister (void) {
	lua_register(L,"Out",Out);
	lua_register(L,"OutFormat",OutFormat);
	lua_register(L,"Frame",Frame);
	lua_register(L,"Log",Log);
	lua_register(L,"TLoad",TLoad);
	lua_register(L,"TCreate",TCreate);
	lua_register(L,"TSave",TSave);
	lua_register(L,"TDelete",TDelete);
	lua_register(L,"TClear",TClear);
	lua_register(L,"TFill",TFill);
	lua_register(L,"TDup",TDup);
	lua_register(L,"TPut",TPut);
	lua_register(L,"TPut2",TPut2);
	lua_register(L,"TInfo",TInfo);
	lua_register(L,"TBlend",TBlend);
	lua_register(L,"VOpen",VOpen);
	lua_register(L,"VGet",VGet);
	lua_register(L,"VClose",VClose);
	lua_register(L,"VInfo",VInfo);
	lua_register(L,"OutInfo",OutInfo);
	lua_register(L,"DLLBind",DLLBind);
	lua_register(L,"GetVersion",GetVersion);
	lua_register(L,"Filter",Filter);
	lua_register(L,"Lag",Lag);
	lua_register(L,"L1",L1_l);
	lua_register(L,"Spl",Spl);
}

int GetVersion(lua_State *L) {
	lua_settop(L,0);
	lua_pushnumber(L,VERSION);
	return 1;
}
int Filter(lua_State *L) {
	if (!lua_isnumber(L,1)) {
		luaL_error(L,"Filter̈ԈĂ܂BFilter(filternumber)");
	}
	filter=lua_tointeger(L,1);
	switch (filter) {
		case 0:
			printf ("tB^OɃjAXglCo[@w肵܂B\n");
			break;
		case 1:
			printf ("tB^OɃoCjA@w肵܂B\n");
			break;
		case 2:
			printf ("tB^OɃoCL[rbNtB^Ow肵܂B\n");
			break;
		case 3:
			printf ("tB^OLanczos2tB^Ow肵܂B\n");
			break;

		default:
			luaL_error(L,"Filter:`ĂȂtB^ԍw肵܂B");
			break;
	}
	lua_settop(L,0);
	return 0;
}
int Lag(lua_State *L) {
	if (!lua_istable(L,1)) {
		luaL_error(L,"Lag̈ԈĂ܂Bp=Lag(xarray,parray,x)");
	}
	if (!lua_istable(L,2)) {
		luaL_error(L,"Lag̈ԈĂ܂Bp=Lag(xarray,parray,x)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"Lag̈ԈĂ܂Bp=Lag(xarray,parray,x)");
	}
	double x=lua_tonumber(L,3);
//	printf ("x=%lf\n",x);

	double xn[64];
	double pn[64];
	int i;
	for (i=0;i<64;i++) {
		//el̎擾
		lua_pushnumber(L,i+1);
		lua_gettable(L,1);
		if (!lua_isnumber(L,-1)) {
			if (i<2) {
				luaL_error(L,"Lag:OWԂɂ͂Q_ȏ̓_^ĂB");
				return 0;
			}
			lua_pop(L,0);
			break;
		}
		xn[i]=lua_tonumber(L,-1);
		lua_pop(L,0);
		lua_pushnumber(L,i+1);
		lua_gettable(L,2);
		if (!lua_isnumber(L,-1)) {
			lua_pop(L,0);
			break;
		}
		pn[i]=lua_tonumber(L,-1);
		lua_pop(L,0);
//		printf ("xn[%d]=%lf pn[%d]=%lf\n",i,xn[i],i,pn[i]);
	}
	//OWԂB
	int n=i;
//	printf ("n=%d\n",n);
	lua_settop(L,0);

	double p=0.0;
	for (i=0;i<n;i++) {
		double bunshi,bunbo;
		//q̌vZ r00ɂB
		bunshi=1.0;
		for (int j=0;j<n;j++) {
			if (i==j) continue;
			double l=x-xn[j];
			if (fabs(l)<0.0000001) {bunshi=0.0;break;}
			bunshi=bunshi*l;
		}
		if (bunshi==0.0 || fabs(bunshi)<0.0000001) continue;
		//̌vZ
		bunbo=1.0;
		for (int j=0;j<n;j++) {
			if (i==j) continue;
			double l=xn[i]-xn[j];
			if (fabs(l)<0.0000001) {bunbo=0.0;break;}
			bunbo=bunbo*l;
		}
		if (bunbo==0.0 || fabs(bunbo)<0.0000001) {
			luaL_error(L,"Lag:OWԂɎs܂Bsȓ_w肵Ă܂B");
			return 0;
		}
		p=p+pn[i]*bunshi/bunbo;
	}
	lua_pushnumber(L,p);
	return 1;
}
void maketable (double x[],double y[],double z[],int n);
void maketable (double x[],double y[],double z[],int n) {
	int i;
	double t;
	double h[64],d[64];
	z[0]=0;z[n-1]=0;
	for (i=0;i<n-1;i++) {
		h[i]=x[i+1]-x[i];
		d[i+1]=(y[i+1]-y[i])/h[i];
	}
	z[1]=d[2]-d[1]-h[0]*z[0];
	d[1]=2*(x[2]-x[0]);
	for (i=1;i<n-2;i++) {
		t=h[i]/d[i];
		z[i+1]=d[i+2]-d[i+1]-z[i]*t;
		d[i+1]=2*(x[i+2]-x[i])-h[i]*t;
	}
	z[n-2]-=h[n-2]*z[n-1];
	for (i=n-2;i>0;i--) {
		z[i]=(z[i]-h[i]*z[i+1])/d[i];
	}
}
double spline (double t,double x[],double y[],double z[],int n);
double spline (double t,double x[],double y[],double z[],int n) {
	int i,j,k;
	double d,h;
	i=0;j=n-1;
	while(i<j) {
		k=(i+j)/2;
		if (x[k]<t) i=k+1; else j=k;
	}
	if (i>0) i--;
	h=x[i+1]-x[i];d=t-x[i];
	return (((z[i+1]-z[i])*d/h+z[i]*3)*d
		+((y[i+1]-y[i])/h
		-(z[i]*2+z[i+1])*h))*d+y[i];
}
int Spl(lua_State *L) {
	if (!lua_istable(L,1)) {
		luaL_error(L,"Spl̈ԈĂ܂Bp=Spl(xarray,parray,x)");
	}
	if (!lua_istable(L,2)) {
		luaL_error(L,"Spl̈ԈĂ܂Bp=Spl(xarray,parray,x)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"Spl̈ԈĂ܂Bp=Spl(xarray,parray,x)");
	}
	double x=lua_tonumber(L,3);
	double xn[64];
	double pn[64];
	int i;
	for (i=0;i<64;i++) {
		//el̎擾
		lua_pushnumber(L,i+1);
		lua_gettable(L,1);
		if (!lua_isnumber(L,-1)) {
			if (i<2) {
				luaL_error(L,"Spl:XvCԂɂ͂Q_ȏ̓_^ĂB");
				return 0;
			}
			lua_pop(L,0);
			break;
		}
		xn[i]=lua_tonumber(L,-1);
		lua_pop(L,0);
		lua_pushnumber(L,i+1);
		lua_gettable(L,2);
		if (!lua_isnumber(L,-1)) {
			lua_pop(L,0);
			break;
		}
		pn[i]=lua_tonumber(L,-1);
		lua_pop(L,0);
	}
	//XvCԂB
	int n=i;
	lua_settop(L,0);

	double zn[64];
	if (n>63) {
		//G[
		luaL_error(L,"SplF_̐64܂łłB");
		return 0;
	}
	for(i=0;i<n;i++) {
		if(xn[i]==x) {
			lua_pushnumber(L,pn[i]);
			return 1;
		}
		if (i<n-1 && xn[i]>=xn[i+1]) {
			//G[
			luaL_error(L,"Spl:xarray͓l܂܂ȂAɃ\[gꂽzłȂ΂Ȃ܂B");
			return 0;
		}
	}
	if (x<xn[0] || x>xn[n-1]) {
		//G[
		luaL_error(L,"SplF߂xW͈͊OłB");
		return 0;
	}
	maketable(xn,pn,zn,n);
	lua_pushnumber(L,spline(x,xn,pn,zn,n));
	return 1;
}

int L1_l(lua_State *L) {
	if (!lua_isnumber(L,1)) {
		luaL_error(L,"L1̈ԈĂ܂Bp=L1(x0,p0,x1,p1,x)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"L1̈ԈĂ܂Bp=L1(x0,p0,x1,p1,x)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"L1̈ԈĂ܂Bp=L1(x0,p0,x1,p1,x)");
	}
	if (!lua_isnumber(L,4)) {
		luaL_error(L,"L1̈ԈĂ܂Bp=L1(x0,p0,x1,p1,x)");
	}
	if (!lua_isnumber(L,5)) {
		luaL_error(L,"L1̈ԈĂ܂Bp=L1(x0,p0,x1,p1,x)");
	}
	double x,x0,p0,x1,p1,p;
	x0=lua_tonumber(L,1);
	p0=lua_tonumber(L,2);
	x1=lua_tonumber(L,3);
	p1=lua_tonumber(L,4);
	x=lua_tonumber(L,5);
	p=L1(x0,p0,x1,p1,x);
	lua_settop(L,0);
	lua_pushnumber(L,p);
	return 1;
}

int Log(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"Log̈ԈĂ܂BLog(text)");
	}
	char log[512];
	sprintf_s(log,512,"OóF%s",lua_tostring(L,1));
	puts (log);
	lua_settop(L,0);
	return 0;
}
int TLoad(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TLoad̈ԈĂ܂BTLoad(texname,filename)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"TLoad̈ԈĂ܂BTLoad(texname,filename)");
	}
	const char *texname=lua_tostring(L,1);
	const char *filename=lua_tostring(L,2);
	DIB *newdib=new DIB;
	if (!newdib->Load(filename)) {
		delete newdib;
		luaL_error(L,"TLoad:t@C\"%s\"J܂B",filename);
	}
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr!=dibmap.end()) {
		//̂ō폜B
		itr->second->Delete();
		delete (itr->second);
		dibmap.erase(itr);
	}
	dibmap.insert(pair<string,DIB *>(texname,newdib));
	lua_settop(L,0);
	return 0;
}
int TCreate(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TCreatëԈĂ܂BTCreate(texname,width,height)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"TCreatëԈĂ܂BTCreate(texname,width,height)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"TCreatëԈĂ܂BTCreate(texname,width,height)");
	}
	const char *texname=lua_tostring(L,1);
	int w=lua_tointeger(L,2);
	int h=lua_tointeger(L,3);
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr!=dibmap.end()) {
		//̂ō폜B
		itr->second->Delete();
		delete (itr->second);
		dibmap.erase(itr);
	}
	DIB *newdib=new DIB;
	dibmap.insert(pair<string,DIB *>(texname,newdib));
	newdib->Create(w,h);
	lua_settop(L,0);
	return 0;
}
int TSave(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TSavëԈĂ܂BTSave(texname,filename)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"TSavëԈĂ܂BTSave(texname,filename)");
	}
	const char *texname=lua_tostring(L,1);
	const char *savename=lua_tostring(L,2);
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TSave:݂ȂeNX`\"%s\"w肵܂B",texname);
		return 0;
	}
	if (!itr->second->Save(savename)) {
		luaL_error(L,"TSave:t@C\"%s\"̃Z[uɎs܂B",savename);
		return 0;
	}
	lua_settop(L,0);
	return 0;
}
int TDelete(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TDeletëԈĂ܂BTDelete(texname)");
	}
	const char *texname=lua_tostring(L,1);
	if (texname[0]=='\0') {
		luaL_error(L,"TDelete:XN[eNX`Ƃ܂B",texname);
		return 0;
	}
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TDelete:݂ȂeNX`\"%s\"폜悤Ƃ܂B",texname);
		return 0;
	}
	//̂ō폜B
	itr->second->Delete();
	delete (itr->second);
	dibmap.erase(itr);
	lua_settop(L,0);
	return 0;
}
int TClear(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TClear̈ԈĂ܂BTClear(texname,color)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"TClear̈ԈĂ܂BTClear(texname,color)");
	}

	const char *texname=lua_tostring(L,1);
	DWORD col=(DWORD)lua_tonumber(L,2);

	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TClear:݂ȂeNX`\"%s\"w肵܂B",texname);
		return 0;
	}
	//̂œhԂB
	itr->second->Clear(col);
	lua_settop(L,0);
	return 0;
}
int TFill(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}
	if (!lua_isnumber(L,4)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}
	if (!lua_isnumber(L,5)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}
	if (!lua_isnumber(L,6)) {
		luaL_error(L,"TFill̈ԈĂ܂BTFill(texname,lx,ly,width,height,color)");
	}

	const char *texname=lua_tostring(L,1);
	int lx=lua_tointeger(L,2);
	int ly=lua_tointeger(L,3);
	int wd=lua_tointeger(L,4);
	int ht=lua_tointeger(L,5);
	unsigned int col=(unsigned int)lua_tonumber(L,6);
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TexFill:݂ȂeNX`\"%s\"w肵܂B",texname);
		return 0;
	}
	itr->second->Fill(lx,ly,wd,ht,col);
	lua_settop(L,0);
	return 0;
}
int TDup(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TDup̈ԈĂ܂BTDup(desttexname,srctexname)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"TDup̈ԈĂ܂BTDup(desttexname,srctexname)");
	}
	const char *destname=lua_tostring(L,1);
	const char *srcname=lua_tostring(L,2);
	DIBmap::iterator ditr=dibmap.find(destname);
	DIBmap::iterator sitr=dibmap.find(srcname);
	if (ditr!=dibmap.end()) {
		//ɑ݂̂ŁAx폜B
		ditr->second->Delete();
		delete ditr->second;
		dibmap.erase(ditr);
	}
	if (sitr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TDup:Rs[̃eNX`\"%s\"݂܂B",srcname);
		return 0;
	}
	DIB *pdib=new DIB;
	pdib->Duplicate(sitr->second);
	dibmap.insert(pair<string,DIB *>(destname,pdib));
	lua_settop(L,0);
	return 0;
}
int TPut(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,4)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,5)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isstring(L,6)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,7)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,8)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}
	if (!lua_isnumber(L,9)) {
		luaL_error(L,"TPuẗԈĂ܂BTPut(desttexname,destx,desty,width,height,srctexname,srcx,srcy,alpha)");
	}

	const char *destname=lua_tostring(L,1);
	int x=lua_tointeger(L,2);
	int y=lua_tointeger(L,3);
	int w=lua_tointeger(L,4);
	int h=lua_tointeger(L,5);
	const char *srcname=lua_tostring(L,6);
	int srcx=lua_tointeger(L,7);
	int srcy=lua_tointeger(L,8);
	int alpha=lua_tointeger(L,9);
	DIBmap::iterator ditr=dibmap.find(destname);
	DIBmap::iterator sitr=dibmap.find(srcname);
	if (ditr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TPut:_ÕeNX`\"%s\"݂܂B",destname);
		return 0;
	}
	if (sitr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TPut:_ÕeNX`\"%s\"݂܂B",srcname);
		return 0;
	}

	//̂Ń_O
	ditr->second->Render(x,y,w,h,sitr->second,srcx,srcy,alpha,blendmode);
	lua_settop(L,0);
	return 0;
}
int TPut2(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isstring(L,4)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,5)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,6)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,7)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,8)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,9)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,10)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,11)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}
	if (!lua_isnumber(L,12)) {
		luaL_error(L,"TPut2̈ԈĂ܂BTPut2(desttexname,destcx,destcy,srctexname,srclx,srcly,srcwidth,srcheight,xscale,yscale,rotation,alpha)");
	}

	const char *destname=lua_tostring(L,1);
	int dcx=lua_tointeger(L,2);
	int dcy=lua_tointeger(L,3);
	const char *srcname=lua_tostring(L,4);
	int slx=lua_tointeger(L,5);
	int sly=lua_tointeger(L,6);
	int sw=lua_tointeger(L,7);
	int sh=lua_tointeger(L,8);
	double xs=lua_tonumber(L,9);
	double ys=lua_tonumber(L,10);
	double rot=lua_tonumber(L,11);
	int alpha=lua_tointeger(L,12);
	DIBmap::iterator ditr=dibmap.find(destname);
	DIBmap::iterator sitr=dibmap.find(srcname);
	if (ditr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TPut2:_ÕeNX`\"%s\"݂܂B",destname);
		return 0;
	}
	if (sitr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TPut2:_ÕeNX`\"%s\"݂܂B",srcname);
		return 0;
	}

	//̂Ń_O
	switch (filter) {
		case 0:
			ditr->second->Render2_0(dcx,dcy,sitr->second,slx,sly,sw,sh,xs,ys,rot,alpha,blendmode);
			break;
		case 1:
			ditr->second->Render2_1(dcx,dcy,sitr->second,slx,sly,sw,sh,xs,ys,rot,alpha,blendmode);
			break;
		case 2:
			ditr->second->Render2_2(dcx,dcy,sitr->second,slx,sly,sw,sh,xs,ys,rot,alpha,blendmode);
			break;
		case 3:
			ditr->second->Render2_3(dcx,dcy,sitr->second,slx,sly,sw,sh,xs,ys,rot,alpha,blendmode);
			break;
	}
	lua_settop(L,0);
	return 0;
}

int TInfo(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TInföԈĂ܂BTInfo(texname)");
	}
	const char *texname=lua_tostring(L,1);
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"TInfo:݂ȂeNX`\"%s\"̏𓾂悤Ƃ܂B",texname);
		return 0;
	}
	//̂ŏԂ
	lua_settop(L,0);
	lua_pushinteger(L,itr->second->width);
	lua_pushinteger(L,itr->second->height);
	return 2;
}

int TBlend(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"TBlend̈ԈĂ܂BTBlend(mode)");
	}
	const char *mode=lua_tostring(L,1);
	char buf[256];
	int i;
	for (i=0;mode[i]!='\0';i++) {
		buf[i]=tolower(mode[i]);
	}
	buf[i]='\0';
	if (strcmp(buf,"alpha")==0 || buf[0]=='\0') {
		blendmode=BM_ALPHA;
	} else if (strcmp(buf,"add")==0) {
		blendmode=BM_ADD;
	} else if (strcmp(buf,"copy")==0) {
		blendmode=BM_COPY;
	} else {
		luaL_error(L,"TBlend:uh[h̐ݒ肪ԈĂ܂B\"%s\"",mode);
	}
	return 0;
}

int VOpen(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"VOpen̈ԈĂ܂BVOpen(videoname,filename)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"VOpen̈ԈĂ܂BVOpen(videoname,filename)");
	}
	const char *videoname=lua_tostring(L,1);
	const char *filename=lua_tostring(L,2);
	AVImap::iterator itr=avimap.find(videoname);
	if (itr!=avimap.end()) {
		//̂ō폜B
		itr->second->Delete();
		delete (itr->second);
		avimap.erase(itr);
	}
	AVI *newavi=new AVI;
	avimap.insert(pair<string,AVI *>(videoname,newavi));
	newavi->Load(filename);
	lua_settop(L,0);
	return 0;
}

int VGet(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"VGeẗԈĂ܂BVGet(texname,videoname,frame)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"VGeẗԈĂ܂BVGet(texname,videoname,frame)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"VGeẗԈĂ܂BVGet(texname,videoname,frame)");
	}
	const char *texname=lua_tostring(L,1);
	const char *videoname=lua_tostring(L,2);
	int frame=lua_tointeger(L,3);
	AVImap::iterator itr=avimap.find(videoname);
	if (itr==avimap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"VGet:݂Ȃ[r[\"%s\"w肵܂B",videoname);
		return 0;
	}
	DIBmap::iterator ditr=dibmap.find(texname);
	if (ditr!=dibmap.end()) {
		//̂ŁA폜B
		ditr->second->Delete();
		delete ditr->second;
		dibmap.erase(ditr);
	}
	
	DIB *newdib=new DIB;

	ditr=(dibmap.insert(pair<string,DIB *>(texname,newdib))).first;
	itr->second->GetFrame(ditr->second,frame);//ō쐬B

	lua_settop(L,0);
	return 0;
}

int VInfo(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"VInföԈĂ܂Bwidth,height,fps,framenum,start=VInfo(videoname)");
	}
	const char *videoname=lua_tostring(L,1);
	AVImap::iterator itr=avimap.find(videoname);
	if (itr==avimap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"VInfo:݂Ȃ[r[\"%s\"̏擾悤Ƃ܂B",videoname);
		return 0;
	}
	lua_settop(L,0);
	AVIFILEINFOW fi;
	AVISTREAMINFOW si;
	itr->second->pavi->Info(&fi,sizeof(AVIFILEINFOW));
	itr->second->pstm->Info(&si,sizeof(AVISTREAMINFOW));
	lua_pushinteger(L,fi.dwWidth);
	lua_pushinteger(L,fi.dwHeight);
	lua_pushnumber(L,si.dwRate/si.dwScale);
	lua_pushinteger(L,AVIStreamLength(itr->second->pstm));
	lua_pushinteger(L,si.dwStart);
	return 5;
}

int VClose(lua_State *L) {
	if (!lua_isstring(L,1)) {
		luaL_error(L,"VClosëԈĂ܂BVClose(videoname)");
	}
	const char *videoname=lua_tostring(L,1);
	AVImap::iterator itr=avimap.find(videoname);
	if (itr==avimap.end()) {
		//Ȃ̂ŁAG[
		luaL_error(L,"VClose:݂Ȃ[r[\"%s\"폜悤Ƃ܂B",videoname);
		return 0;
	}
	itr->second->Delete();
	delete (itr->second);
	avimap.erase(itr);
	lua_settop(L,0);
	return 0;
}

BITMAPINFO *outbi=NULL;//t[ɏDIBBITMAPINFÕ|C^
unsigned char *outb=NULL;//t[ɏDIBlpbits̃|C^

int OutInfo(lua_State *L) {
	if (!startflag) {
		luaL_error(L,"OutInfo:܂o̓t@CJĂ܂B");
	}
	lua_settop(L,0);
	lua_pushinteger(L,width);
	lua_pushinteger(L,height);
	lua_pushinteger(L,fps);
	lua_pushinteger(L,framenum);
	return 4;
}

int OutFormat(lua_State *L) {
	outformat=true;
	lua_settop(L,0);
	return 0;
}

int Out(lua_State *L) {
	BITMAPINFOHEADER obi;
	if (!lua_isstring(L,1)) {
		luaL_error(L,"OuẗԈĂ܂BOut(filename,width,height,fps,framenum)");
	}
	if (!lua_isnumber(L,2)) {
		luaL_error(L,"OuẗԈĂ܂BOut(filename,width,height,fps,framenum)");
	}
	if (!lua_isnumber(L,3)) {
		luaL_error(L,"OuẗԈĂ܂BOut(filename,width,height,fps,framenum)");
	}
	if (!lua_isnumber(L,4)) {
		luaL_error(L,"OuẗԈĂ܂BOut(filename,width,height,fps,framenum)");
	}
	if (!lua_isnumber(L,5)) {
		luaL_error(L,"OuẗԈĂ܂BOut(filename,width,height,fps,framenum)");
	}
	width=lua_tointeger(L,2);
	height=lua_tointeger(L,3);
	fps=lua_tointeger(L,4);
	obi.biBitCount=32;
	obi.biClrImportant=0;
	obi.biClrUsed=0;
	obi.biCompression=BI_RGB;
	obi.biHeight=height;
	obi.biPlanes=1;
	obi.biWidth=width;
	obi.biXPelsPerMeter=0;
	obi.biYPelsPerMeter=0;
	obi.biSizeImage=width*height*4;
	obi.biSize=sizeof(BITMAPINFOHEADER);
	outbi=(BITMAPINFO *)malloc(obi.biSizeImage+sizeof(BITMAPINFOHEADER));
	outbi->bmiHeader=obi;
	outb=(unsigned char *)outbi->bmiColors;
	framenum=lua_tointeger(L,5);
	const char *filename=lua_tostring(L,1);
	WriteStart(filename,&obi,fps,framenum);
	printf("AVIt@CݒF\nt@C\"%s\" F%d F%d\nb%dt[ t[F%d\n",filename,width,height,fps,framenum);
	
	lua_settop(L,0);
	return 0;
}

int Frame(lua_State *L) {
	if (!lua_isnumber(L,1)) {
		luaL_error(L,"FramëԈĂ܂BFrame(frame,texname)");
	}
	if (!lua_isstring(L,2)) {
		luaL_error(L,"FramëԈĂ܂BFrame(frame,texname)");
	}
	int frame=lua_tointeger(L,1);
	const char *texname=lua_tostring(L,2);
	DIBmap::iterator itr=dibmap.find(texname);
	if (itr==dibmap.end()) {
		luaL_error(L,"݂ȂeNX`w肵܂F%s",texname);
	}
	DIB *pd=itr->second;
	if (framenum<=frame || 0>frame) {
		luaL_error(L,"͈͊Õt[%dw肵܂Bt[ԍ0`%dłB",frame,framenum-1);
	}
	if (pd->width!=width || pd->height!=height) {
		luaL_error(L,"o͂Ɏw肵eNX`̃TCYAVIt@C̉ʃTCYƈႢ܂B");
	}
	Write(frame,pd->lpbits);
	printf ("\r(%d/%d)\r",frame,framenum-1);
	lua_settop(L,0);
	return 0;
}

void WriteStart (const char *filename,LPBITMAPINFOHEADER lpbi,int fps,int fnum) {
	if (startflag) {
		WriteEnd();
		startflag=false;
	}
	int width=lpbi->biWidth;
	int height=lpbi->biHeight;

	if (outformat) {
		COMPVARS compvars;
		PAVISTREAM pstmt;
		memset(&compvars,0,sizeof(compvars));
		compvars.cbSize=sizeof(compvars);
		if (!ICCompressorChoose(NULL, 0, NULL, NULL, &compvars, NULL)) {
			luaL_error(L,"tH[}bg̎wɎs܂B");
			return;
		}
		if (compvars.fccHandler==comptypeDIB) goto _notcompress;
		if (AVIFileOpen(&pavi,filename,OF_CREATE | OF_WRITE | OF_SHARE_DENY_NONE,NULL)!=0) {
			luaL_error(L,"o̓t@CJ܂B");
			return;
		}
		memset(&si,0,sizeof(AVISTREAMINFO));
		si.fccType=streamtypeVIDEO;
		si.fccHandler=compvars.fccHandler;
		si.dwScale=1;
		si.dwRate=fps;
		si.rcFrame.left=0;
		si.rcFrame.top=0;
		si.rcFrame.right=width;
		si.rcFrame.bottom=height;
		if (AVIFileCreateStream(pavi,&pstmt,&si)!=0) {
			luaL_error(L,"w肳ꂽtH[}bgŃXg[J܂łB");
			AVIFileRelease(pavi);
			return;
		}
		AVICOMPRESSOPTIONS options;
		options.fccType=streamtypeVIDEO;
		options.fccHandler=compvars.fccHandler;
		options.dwKeyFrameEvery=compvars.lKey;
		options.dwQuality=compvars.lQ;
		options.dwBytesPerSecond=compvars.lDataRate;
		options.dwFlags=AVICOMPRESSF_DATARATE | AVICOMPRESSF_KEYFRAMES;
		options.lpFormat=NULL;
		options.cbFormat=0;
		options.lpParms=NULL;
		options.cbParms=0;
		options.dwInterleaveEvery=0;
		if (AVIMakeCompressedStream(&pstm, pstmt, &options, NULL) != AVIERR_OK) {
			AVIStreamClose(pstm);
			AVIFileRelease(pavi); 
			luaL_error(L,"w肳ꂽtH[}bgňkXg[J܂łB");
			return;
		}
		AVIStreamSetFormat(pstm,0,lpbi,sizeof(BITMAPINFOHEADER));
		AVIStreamClose(pstmt);
		startflag=true;
		printf("AVIt@C\"%s\"o͗pɊJ܂ikďo͂܂jB\n",filename);
		return;
	} else {
_notcompress:
		si.dwCaps=0;
		si.dwEditCount=0;
		si.dwFlags=0;
		si.dwFormatChangeCount=0;
		si.dwInitialFrames=0;
		si.dwLength=fnum;
		si.dwQuality=(DWORD)-1;
		si.dwRate=fps;
		si.dwScale=1;
		si.dwSampleSize=0;
		si.dwScale=1;
		si.dwStart=0;
		si.dwSuggestedBufferSize=0;
		si.fccHandler=comptypeDIB;
		si.fccType=streamtypeVIDEO;
		si.rcFrame.left=0;
		si.rcFrame.top=0;
		si.rcFrame.right=width;
		si.rcFrame.bottom=height;
		strcpy_s(si.szName,64,"Video #1");
		si.wLanguage=0;
		si.wPriority=0;
		sizeimage=lpbi->biSizeImage;
		if (AVIFileOpen(&pavi,filename,OF_CREATE | OF_WRITE | OF_SHARE_DENY_NONE,NULL)!=0) {
			luaL_error(L,"o̓t@CJ܂B");
			return;
		}
		if (AVIFileCreateStream(pavi,&pstm,&si)!=0) {
			luaL_error(L,"o̓Xg[܂B");
			return;
		}
		if (AVIStreamSetFormat(pstm,0,lpbi,sizeof(BITMAPINFOHEADER))!=0) {
			luaL_error(L,"o̓tH[}bgwł܂B");
			return;
		}
		startflag=true;
		printf("AVIt@C\"%s\"o͗pɊJ܂ikw肳Ă܂jB\n",filename);
		return;
	}
}

void Write (int frame,unsigned char *lpbits) {
	if (!startflag) {
		luaL_error(L,"o̓t@Cw肳Ă܂B");
		return;
	}
	int y=width*(height-1)*4;
	int o=0;
	int ofs=width*4;
	for (int i=0;i<height;i++) {
		memcpy(outb+o,lpbits+y,width*4);
		y-=ofs;
		o+=ofs;
	}
	if (AVIStreamWrite(pstm,frame,1,outb,sizeimage,AVIIF_KEYFRAME,NULL,NULL)!=0) {
		luaL_error(L,"Xg[ւ݂̏s܂B");
		return;
	}
	return;
}

void WriteEnd (void) {
	if (startflag) {
		AVIStreamRelease(pstm);
		AVIFileRelease(pavi);
		printf("o̓t@C܂B\n");
	}
	return;
}

int main (int argc,char **argv) {
	blendmode=BM_ALPHA;
	outformat=true;
	filter=3;//ftHgLanczos2
	AVIFileInit();
	printf ("\n------------------\n");
	printf ("Free Effect Script Ver%1.2lf sO\n\n",VERSION);
	L=lua_open();
	luaL_openlibs(L);
	LuaRegister();
	if (argc<2) {
		printf ("XNvgt@Cw肵ĂB\n");
	} else {
		char drv[256],dir[256],fnm[256],ext[256];
		_splitpath_s(argv[1],drv,256,dir,256,fnm,256,ext,256);
		char buf[256];
		sprintf_s (buf,256,"%s%s",drv,dir);
		_chdir(buf);
		sprintf_s (buf,256,"%s%s",fnm,ext);
		printf ("t@C\"%s\"s܂B\n",buf);
		int ret=luaL_dofile(L,argv[1]);
		if (ret!=0) {
			printf("XNvgG[܂B\n%s\n",lua_tostring(L,-1));
			lua_pop(L,1);
		} else {
			printf("XNvg͐I܂B\n");
		}
	}
	//DIBAVǏn
	DIBmap::iterator itrend=dibmap.end();
	DIBmap::iterator itr;
	for(itr=dibmap.begin();itr!=itrend;itr++) {
		DIB *pdib=itr->second;
		pdib->Delete();
		delete pdib;
	}
	dibmap.clear();
	AVImap::iterator aitend=avimap.end();
	AVImap::iterator ait;
	for (ait=avimap.begin();ait!=aitend;ait++) {
		AVI *pavi=ait->second;
		pavi->Delete();
		delete pavi;
	}
	avimap.clear();
	if (outbi!=NULL) free(outbi);
	WriteEnd();
	printf ("------------------\n\n");
	if (_isatty(_fileno(stdout))) system("pause");
	lua_close(L);
	AVIFileExit();
}

int DLLBind(lua_State *L) {
	if (!lua_isstring(L,1) || !lua_isstring(L,2) || !lua_isstring(L,3)) {
		luaL_error(L,"DLLBind̈ԈĂ܂BDllBind(dllname,dllfuncname,luafuncname)");
	}
	const char *dllname,*dllfuncname,*luafuncname;
	dllname=lua_tostring(L,1);
	dllfuncname=lua_tostring(L,2);
	luafuncname=lua_tostring(L,3);
	int (*pf)(lua_State *);
	pf=(int (*)(lua_State *))dm.Get(dllname,dllfuncname);
	if (pf==NULL) {
		luaL_error(L,"DLLBindFw肳ꂽDLL\"%s\"֐\"%s\"ǂݍ߂܂B",dllname,dllfuncname);
	}
	lua_register(L,luafuncname,pf);
	lua_settop(L,0);
	return 0;
}

void DIB::Render2_1 (
	int dcx,int dcy,
	DIB *src,int slx,int sly,int sw,int sh,
	double xs,double ys,double rot,
	int srcalpha,BlendMode md) {
	if (srcalpha<=0) return;
	if (srcalpha>255) srcalpha=255;
	if (fabs(xs-1.0)<0.0000001 && fabs(ys-1.0)<0.0000001 && fabs(rot)<0.0000001) {
		Render(dcx-sw/2,dcy-sh/2,sw,sh,src,slx,sly,srcalpha,md);
		return;
	}
	matrix m;
	//ꎟϊvZB
	if (xs==0 || ys==0) return;//ׂĂ
	double rotrad=-rot*3.1415926536/180.0;
	//lvZB
	//ϊsłB
	m._00=cos(rotrad)*xs;m._10=-sin(rotrad)*xs;
	m._01=sin(rotrad)*ys;m._11=cos(rotrad)*ys;

	double ox0,oy0,ox1,oy1,ox2,oy2,ox3,oy3;
	double sdw=double(sw),sdh=double(sh);
	mulmatrix(&ox0,&oy0,-sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox1,&oy1,+sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox2,&oy2,-sdw/2.0,-sdh/2.0,&m);
	mulmatrix(&ox3,&oy3,+sdw/2.0,-sdh/2.0,&m);
	oy0=-oy0;
	oy1=-oy1;
	oy2=-oy2;
	oy3=-oy3;
	int dlx,dly,dw,dh;
	int oix0,oix1,oix2,oix3;
	int oiy0,oiy1,oiy2,oiy3;
	oix0=int(ox0);oiy0=int(oy0);
	oix1=int(ox1);oiy1=int(oy1);
	oix2=int(ox2);oiy2=int(oy2);
	oix3=int(ox3);oiy3=int(oy3);
	dlx=min(min(min(oix0,oix1),oix2),oix3);
	dly=min(min(min(oiy0,oiy1),oiy2),oiy3);
	dw=max(max(max(oix0,oix1),oix2),oix3);
	dh=max(max(max(oiy0,oiy1),oiy2),oiy3);
	dw=dw-dlx;
	dh=dh-dly;
	dw=(int)(sqrt(double(dw*dw+dh*dh))+2);
	dh=dw;
	dlx=-dw/2;
	dly=-dw/2;
	//ŁAXLC͈͔͂B
	//dest̋`ɒB
	//tϊsłBY]ς
	m._00=cos(-rotrad)/xs;m._10=sin(-rotrad)/ys;
	m._01=-sin(-rotrad)/xs;m._11=cos(-rotrad)/ys;
	double dx,dy;
	double x,y;
	double ox,oy;
	mulmatrix(&dx,&dy,dlx+0.5,dly+0.5,&m);
	x=dx;
	y=dy;
	int s_b,s_g,s_r,s_a;
	unsigned char *srcb=src->lpbits;
	unsigned char *destb=lpbits;
	int ptr=(dlx+dcx)*4+(dly+dcy)*width*4;
	int dp=0;
	int srcpitch=src->width*4;
	int destpitch=width*4;
	s_b=s_g=s_r=s_a=0;
	for (int j=0;j<dh;j++) {
		ox=x;oy=y;
		for (int i=0;i<dw;i++) {
			if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
				goto _next;
			}
			int xa[2],ya[2],pa[2][2];
			if (x>0) {
				xa[0]=int(x+0.00000000001);
				xa[1]=int(x+0.99999999999);
			} else {
				xa[0]=int(x-0.00000000001);
				xa[1]=int(x-0.99999999999);
			}
			if (y>0) {
				ya[0]=int(y+0.00000000001);
				ya[1]=int(y+0.99999999999);
			} else {
				ya[0]=int(y-0.00000000001);
				ya[1]=int(y-0.99999999999);
			}

			//ߖTS_m
			bool fl=false;
			for (int jj=0;jj<2;jj++) {
				for (int ii=0;ii<2;ii++) {
					if (xa[ii]+sw/2<0 || xa[ii]+sw/2>=sw || ya[jj]+sh/2<0 || ya[jj]+sh/2>=sh) {
						pa[ii][jj]=-1;
					} else {
						fl=true;
						pa[ii][jj]=(xa[ii]+sw/2)*4+(ya[jj]+sh/2)*srcpitch;
					}
				}
			}
			if (!fl) goto _next;
			double p0,p1;
			int c;
			int v00,v01,v10,v11,v;
			v00=(pa[0][0]!=-1)?srcb[pa[0][0]]:0;
			v10=(pa[1][0]!=-1)?srcb[pa[1][0]]:0;
			v01=(pa[0][1]!=-1)?srcb[pa[0][1]]:0;
			v11=(pa[1][1]!=-1)?srcb[pa[1][1]]:0;
			p0=L1(xa[0],v00,xa[1],v10,x);
			p1=L1(xa[0],v01,xa[1],v11,x);
			s_b=(int)(L1(ya[0],p0,ya[1],p1,y));
			v00=(pa[0][0]!=-1)?srcb[pa[0][0]+1]:0;
			v10=(pa[1][0]!=-1)?srcb[pa[1][0]+1]:0;
			v01=(pa[0][1]!=-1)?srcb[pa[0][1]+1]:0;
			v11=(pa[1][1]!=-1)?srcb[pa[1][1]+1]:0;
			p0=L1(xa[0],v00,xa[1],v10,x);
			p1=L1(xa[0],v01,xa[1],v11,x);
			s_g=(int)(L1(ya[0],p0,ya[1],p1,y));
			v00=(pa[0][0]!=-1)?srcb[pa[0][0]+2]:0;
			v10=(pa[1][0]!=-1)?srcb[pa[1][0]+2]:0;
			v01=(pa[0][1]!=-1)?srcb[pa[0][1]+2]:0;
			v11=(pa[1][1]!=-1)?srcb[pa[1][1]+2]:0;
			p0=L1(xa[0],v00,xa[1],v10,x);
			p1=L1(xa[0],v01,xa[1],v11,x);
			s_r=(int)(L1(ya[0],p0,ya[1],p1,y));
			v00=(pa[0][0]!=-1)?srcb[pa[0][0]+3]:0;
			v10=(pa[1][0]!=-1)?srcb[pa[1][0]+3]:0;
			v01=(pa[0][1]!=-1)?srcb[pa[0][1]+3]:0;
			v11=(pa[1][1]!=-1)?srcb[pa[1][1]+3]:0;
			p0=L1(xa[0],v00,xa[1],v10,x);
			p1=L1(xa[0],v01,xa[1],v11,x);
			v=(int)(L1(ya[0],p0,ya[1],p1,y));
			if (v<0) v=0; else if (v>255) v=255;
			s_a=srcalpha*v/255;
			if (s_a==0) goto _next;
			switch (md) {
				case BM_ALPHA:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*s_a/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*s_a/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*s_a/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
				case BM_ADD:
					c=destb[ptr+dp]+s_b;
					destb[ptr+dp]=(c>255)?255:c;
					c=destb[ptr+dp+1]+s_g;
					destb[ptr+dp+1]=(c>255)?255:c;
					c=destb[ptr+dp+2]+s_r;
					destb[ptr+dp+2]=(c>255)?255:c;
					destb[ptr+dp+3]=255;
					break;
				case BM_COPY:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*srcalpha/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*srcalpha/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*srcalpha/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
			}
_next:
			dp+=4;
			x+=m._00;y+=m._10;
		}
		x=ox+m._01;y=oy+m._11;
		dp=0;
		ptr+=destpitch;
	}
}

void DIB::Render2_0 (
	int dcx,int dcy,
	DIB *src,int slx,int sly,int sw,int sh,
	double xs,double ys,double rot,
	int srcalpha,BlendMode md) {
	if (srcalpha<=0) return;
	if (srcalpha>255) srcalpha=255;
	if (fabs(xs-1.0)<0.0000001 && fabs(ys-1.0)<0.0000001 && fabs(rot)<0.0000001) {
		Render(dcx-sw/2,dcy-sh/2,sw,sh,src,slx,sly,srcalpha,md);
		return;
	}
	typedef int INTDOUBLE;
	matrix m;
	//ꎟϊvZB
	if (xs==0 || ys==0) return;//ׂĂ
	double rotrad=-rot*3.1415926536/180.0;
	//lvZB
	//ϊsłB
	m._00=cos(rotrad)*xs;m._10=-sin(rotrad)*xs;
	m._01=sin(rotrad)*ys;m._11=cos(rotrad)*ys;
	double ox0,oy0,ox1,oy1,ox2,oy2,ox3,oy3;
	double sdw=double(sw),sdh=double(sh);
	mulmatrix(&ox0,&oy0,-sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox1,&oy1,+sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox2,&oy2,-sdw/2.0,-sdh/2.0,&m);
	mulmatrix(&ox3,&oy3,+sdw/2.0,-sdh/2.0,&m);
	oy0=-oy0;
	oy1=-oy1;
	oy2=-oy2;
	oy3=-oy3;
	int dlx,dly,dw,dh;
	int oix0,oix1,oix2,oix3;
	int oiy0,oiy1,oiy2,oiy3;
	oix0=int(ox0);oiy0=int(oy0);
	oix1=int(ox1);oiy1=int(oy1);
	oix2=int(ox2);oiy2=int(oy2);
	oix3=int(ox3);oiy3=int(oy3);
	dlx=min(min(min(oix0,oix1),oix2),oix3);
	dly=min(min(min(oiy0,oiy1),oiy2),oiy3);
	dw=max(max(max(oix0,oix1),oix2),oix3);
	dh=max(max(max(oiy0,oiy1),oiy2),oiy3);
	dw=dw-dlx;
	dh=dh-dly;
	dw=(int)(sqrt(double(dw*dw+dh*dh))+2);
	dh=dw;
	dlx=-dw/2;
	dly=-dw/2;
	//ŁAXLC͈͔͂B
	//dest̋`ɒB
	//tϊsłBY]ς
	m._00=cos(-rotrad)/xs;m._10=sin(-rotrad)/ys;
	m._01=-sin(-rotrad)/xs;m._11=cos(-rotrad)/ys;
	INTDOUBLE x,y;
	double dx,dy;
	mulmatrix(&dx,&dy,dlx+0.5,dly+0.5,&m);
	x=int(dx*65536.0);
	y=int(dy*65536.0);
	INTDOUBLE dxx=(INTDOUBLE)(m._00*65536.0);
	INTDOUBLE dxy=(INTDOUBLE)(m._10*65536.0);
	INTDOUBLE dyx=(INTDOUBLE)(m._01*65536.0);
	INTDOUBLE dyy=(INTDOUBLE)(m._11*65536.0);
	//ꂼꑝB
	//قǂ̋`ŕϊsrcWvZAsNZJ[擾ĂB
	//̈O`FbNKvB
	//fXgdcx,dcy \[Xsw sĥYꂸɁB
	//INTDOUBLE65536{̌Œ菬_B
	INTDOUBLE ox,oy;
	int osx=-100000,osy=-100000;
	int s_b,s_g,s_r,s_a;
	unsigned char *srcb=src->lpbits;
	unsigned char *destb=lpbits;
	int ptr=(dlx+dcx)*4+(dly+dcy)*width*4;
	int p;
	int dp=0;
	int srcpitch=src->width*4;
	int destpitch=width*4;
	s_b=s_g=s_r=s_a=0;
	int c;
	if (md==BM_ALPHA) {
		for (int j=0;j<dh;j++) {
			ox=x;oy=y;
			for (int i=0;i<dw;i++) {
				if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
					osx=-10000;
					goto _next;
				}
				int xx=x/65536,yy=y/65536;
				if (xx+sw/2<0 || xx+sw/2>=sw || yy+sh/2<0 || yy+sh/2>=sh) {
					goto _next;
				}
				if (osx!=xx || osy!=yy) {
					//Wς
					osx=xx;
					osy=yy;
					p=(xx+sw/2)*4+(yy+sh/2)*srcpitch;
					s_a=srcalpha*srcb[p+3]/255;
					s_b=srcb[p];
					s_g=srcb[p+1];
					s_r=srcb[p+2];
				}
				destb[ptr+dp]=(s_b-destb[ptr+dp])*s_a/255+destb[ptr+dp];
				destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*s_a/255+destb[ptr+dp+1];
				destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*s_a/255+destb[ptr+dp+2];
				destb[ptr+dp+3]=255;
	_next:
				dp+=4;
				x+=dxx;y+=dxy;
			}
			x=ox+dyx;y=oy+dyy;
			dp=0;
			ptr+=destpitch;
		}
	} else if (md==BM_ADD) {
		for (int j=0;j<dh;j++) {
			ox=x;oy=y;
			for (int i=0;i<dw;i++) {
				if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
					osx=-10000;
					goto _next_1;
				}
				int xx=x/65536,yy=y/65536;
				if (xx+sw/2<0 || xx+sw/2>=sw || yy+sh/2<0 || yy+sh/2>=sh) {
					goto _next_1;
				}
				if (osx!=xx || osy!=yy) {
					//Wς
					osx=xx;
					osy=yy;
					p=(xx+sw/2)*4+(yy+sh/2)*srcpitch;
					s_a=srcalpha*srcb[p+3]/255; 
					s_b=srcb[p]*s_a/255;s_g=srcb[p+1]*s_a/255;s_r=srcb[p+2]*s_a/255;
				}
				c=destb[ptr+dp]+s_b;
				destb[ptr+dp]=(c>255)?255:c;
				c=destb[ptr+dp+1]+s_g;
				destb[ptr+dp+1]=(c>255)?255:c;
				c=destb[ptr+dp+2]+s_r;
				destb[ptr+dp+2]=(c>255)?255:c;
				destb[ptr+dp+3]=255;
	_next_1:
				dp+=4;
				x+=dxx;y+=dxy;
			}
			x=ox+dyx;y=oy+dyy;
			dp=0;
			ptr+=destpitch;
		}
	} else {
		s_a=srcalpha;
		for (int j=0;j<dh;j++) {
			ox=x;oy=y;
			for (int i=0;i<dw;i++) {
				if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
					osx=-10000;
					goto _next_2;
				}
				int xx=x/65536,yy=y/65536;
				if (xx+sw/2<0 || xx+sw/2>=sw || yy+sh/2<0 || yy+sh/2>=sh) {
					goto _next_2;
				}
				if (osx!=xx || osy!=yy) {
					//Wς
					osx=xx;
					osy=yy;
					p=(xx+sw/2)*4+(yy+sh/2)*srcpitch;
					s_b=srcb[p];s_g=srcb[p+1];s_r=srcb[p+2];
				}
				destb[ptr+dp]=(s_b-destb[ptr+dp])*s_a/255+destb[ptr+dp];
				destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*s_a/255+destb[ptr+dp+1];
				destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*s_a/255+destb[ptr+dp+2];
				destb[ptr+dp+3]=255;
	_next_2:
				dp+=4;
				x+=dxx;y+=dxy;
			}
			x=ox+dyx;y=oy+dyy;
			dp=0;
			ptr+=destpitch;
		}
	}
}

void DIB::Render2_2 (
	int dcx,int dcy,
	DIB *src,int slx,int sly,int sw,int sh,
	double xs,double ys,double rot,
	int srcalpha,BlendMode md) {
	if (srcalpha<=0) return;
	if (srcalpha>255) srcalpha=255;
	if (fabs(xs-1.0)<0.0000001 && fabs(ys-1.0)<0.0000001 && fabs(rot)<0.0000001) {
		Render(dcx-sw/2,dcy-sh/2,sw,sh,src,slx,sly,srcalpha,md);
		return;
	}
	matrix m;
	//ꎟϊvZB
	if (xs==0 || ys==0) return;//ׂĂ
	double rotrad=-rot*3.1415926536/180.0;
	//lvZB
	//ϊsłB
	m._00=cos(rotrad)*xs;m._10=-sin(rotrad)*xs;
	m._01=sin(rotrad)*ys;m._11=cos(rotrad)*ys;

	double ox0,oy0,ox1,oy1,ox2,oy2,ox3,oy3;
	double sdw=double(sw),sdh=double(sh);
	mulmatrix(&ox0,&oy0,-sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox1,&oy1,+sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox2,&oy2,-sdw/2.0,-sdh/2.0,&m);
	mulmatrix(&ox3,&oy3,+sdw/2.0,-sdh/2.0,&m);
	oy0=-oy0;
	oy1=-oy1;
	oy2=-oy2;
	oy3=-oy3;
	int dlx,dly,dw,dh;
	int oix0,oix1,oix2,oix3;
	int oiy0,oiy1,oiy2,oiy3;
	oix0=int(ox0);oiy0=int(oy0);
	oix1=int(ox1);oiy1=int(oy1);
	oix2=int(ox2);oiy2=int(oy2);
	oix3=int(ox3);oiy3=int(oy3);
	dlx=min(min(min(oix0,oix1),oix2),oix3);
	dly=min(min(min(oiy0,oiy1),oiy2),oiy3);
	dw=max(max(max(oix0,oix1),oix2),oix3);
	dh=max(max(max(oiy0,oiy1),oiy2),oiy3);
	dw=dw-dlx;
	dh=dh-dly;
	dw=(int)(sqrt(double(dw*dw+dh*dh))+2);
	dh=dw;
	dlx=-dw/2;
	dly=-dw/2;
	//ŁAXLC͈͔͂B
	//dest̋`ɒB
	//tϊsłBY]ς
	m._00=cos(-rotrad)/xs;m._10=sin(-rotrad)/ys;
	m._01=-sin(-rotrad)/xs;m._11=cos(-rotrad)/ys;
	double dx,dy;
	double x,y;
	double ox,oy;
	mulmatrix(&dx,&dy,dlx+0.5,dly+0.5,&m);
	x=dx;
	y=dy;
	int s_b,s_g,s_r,s_a;
	unsigned char *srcb=src->lpbits;
	unsigned char *destb=lpbits;
	int ptr=(dlx+dcx)*4+(dly+dcy)*width*4;
	int dp=0;
	int srcpitch=src->width*4;
	int destpitch=width*4;
	s_b=s_g=s_r=s_a=0;
	for (int j=0;j<dh;j++) {
		ox=x;oy=y;
		for (int i=0;i<dw;i++) {
			if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
				goto _next;
			}
			int xa[4],ya[4],pa[4][4];
			if (x>0) {
				xa[0]=int(x+0.0000001)-1;
				xa[1]=int(x+0.0000001);
				xa[2]=int(x+0.9999999);
				xa[3]=int(x+0.9999999)+1;
			} else {
				xa[0]=int(x-0.0000001)-1;
				xa[1]=int(x-0.0000001);
				xa[2]=int(x-0.9999999);
				xa[3]=int(x-0.9999999)+1;
			}
			if (y>0) {
				ya[0]=int(y+0.0000001)-1;
				ya[1]=int(y+0.0000001);
				ya[2]=int(y+0.9999999);
				ya[3]=int(y+0.9999999)+1;
			} else {
				ya[0]=int(y-0.0000001)-1;
				ya[1]=int(y-0.0000001);
				ya[2]=int(y-0.9999999);
				ya[3]=int(y-0.9999999)+1;
			}
			//ߖTPU_m
			bool fl=false;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					if (xa[ii]+sw/2<0 || xa[ii]+sw/2>=sw || ya[jj]+sh/2<0 || ya[jj]+sh/2>=sh) {
						pa[ii][jj]=-1;
					} else {
						fl=true;
						pa[ii][jj]=(xa[ii]+sw/2)*4+(ya[jj]+sh/2)*srcpitch;
					}
				}
			}
			if (!fl) goto _next;
			double p0,p1,p2,p3,d[4][4];
			p0=p1=p2=p3=0.0;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					double dx=double(xa[ii])-x;
					double dy=double(ya[jj])-y;
					d[ii][jj]=sqrt(dx*dx+dy*dy);
				}
			}
			double tw=0.0;
			double w;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					double dd=d[ii][jj];
					if (dd<1.0) {
						w=1.0-2.0*dd*dd+dd*dd*dd;
					} else if (dd<2.0) {
						w=4.0-8.0*dd+5.0*dd*dd-dd*dd*dd;
					} else continue;
					tw+=w;
					if (pa[ii][jj]==-1) continue;
					p0+=srcb[pa[ii][jj]]*w;
					p1+=srcb[pa[ii][jj]+1]*w;
					p2+=srcb[pa[ii][jj]+2]*w;
					p3+=srcb[pa[ii][jj]+3]*w;
				}
			}
			p0/=tw;
			p1/=tw;
			p2/=tw;
			p3/=tw;
			int p=(int)p3;
			if (p<0) p=0; else if (p>255) p=255;
			s_a=srcalpha*p/255;
			s_b=(int)p0;
			if (s_b<0) s_b=0; else if (s_b>255) s_b=255;
			s_g=(int)p1;
			if (s_g<0) s_g=0; else if (s_g>255) s_g=255;
			s_r=(int)p2;
			if (s_r<0) s_r=0; else if (s_r>255) s_r=255;
			int c;
			switch (md) {
				case BM_ALPHA:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*s_a/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*s_a/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*s_a/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
				case BM_ADD:
					c=destb[ptr+dp]+s_b;
					destb[ptr+dp]=(c>255)?255:c;
					c=destb[ptr+dp+1]+s_g;
					destb[ptr+dp+1]=(c>255)?255:c;
					c=destb[ptr+dp+2]+s_r;
					destb[ptr+dp+2]=(c>255)?255:c;
					destb[ptr+dp+3]=255;
					break;
				case BM_COPY:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*srcalpha/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*srcalpha/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*srcalpha/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
			}
_next:
			dp+=4;
			x+=m._00;y+=m._10;
		}
		x=ox+m._01;y=oy+m._11;
		dp=0;
		ptr+=destpitch;
	}
}

void DIB::Render2_3 (
	int dcx,int dcy,
	DIB *src,int slx,int sly,int sw,int sh,
	double xs,double ys,double rot,
	int srcalpha,BlendMode md) {
	if (srcalpha<=0) return;
	if (srcalpha>255) srcalpha=255;
	if (fabs(xs-1.0)<0.0000001 && fabs(ys-1.0)<0.0000001 && fabs(rot)<0.0000001) {
		Render(dcx-sw/2,dcy-sh/2,sw,sh,src,slx,sly,srcalpha,md);
		return;
	}
	matrix m;
	//ꎟϊvZB
	if (xs==0 || ys==0) return;//ׂĂ
	double rotrad=-rot*3.1415926536/180.0;
	//lvZB
	//ϊsłB
	m._00=cos(rotrad)*xs;m._10=-sin(rotrad)*xs;
	m._01=sin(rotrad)*ys;m._11=cos(rotrad)*ys;

	double ox0,oy0,ox1,oy1,ox2,oy2,ox3,oy3;
	double sdw=double(sw),sdh=double(sh);
	mulmatrix(&ox0,&oy0,-sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox1,&oy1,+sdw/2.0,sdh/2.0,&m);
	mulmatrix(&ox2,&oy2,-sdw/2.0,-sdh/2.0,&m);
	mulmatrix(&ox3,&oy3,+sdw/2.0,-sdh/2.0,&m);
	oy0=-oy0;
	oy1=-oy1;
	oy2=-oy2;
	oy3=-oy3;
	int dlx,dly,dw,dh;
	int oix0,oix1,oix2,oix3;
	int oiy0,oiy1,oiy2,oiy3;
	oix0=int(ox0);oiy0=int(oy0);
	oix1=int(ox1);oiy1=int(oy1);
	oix2=int(ox2);oiy2=int(oy2);
	oix3=int(ox3);oiy3=int(oy3);
	dlx=min(min(min(oix0,oix1),oix2),oix3);
	dly=min(min(min(oiy0,oiy1),oiy2),oiy3);
	dw=max(max(max(oix0,oix1),oix2),oix3);
	dh=max(max(max(oiy0,oiy1),oiy2),oiy3);
	dw=dw-dlx;
	dh=dh-dly;
	dw=(int)(sqrt(double(dw*dw+dh*dh))+2);
	dh=dw;
	dlx=-dw/2;
	dly=-dw/2;
	//ŁAXLC͈͔͂B
	//dest̋`ɒB
	//tϊsłBY]ς
	m._00=cos(-rotrad)/xs;m._10=sin(-rotrad)/ys;
	m._01=-sin(-rotrad)/xs;m._11=cos(-rotrad)/ys;
	double dx,dy;
	double x,y;
	double ox,oy;
	mulmatrix(&dx,&dy,dlx+0.5,dly+0.5,&m);
	x=dx;
	y=dy;
	int s_b,s_g,s_r,s_a;
	unsigned char *srcb=src->lpbits;
	unsigned char *destb=lpbits;
	int ptr=(dlx+dcx)*4+(dly+dcy)*width*4;
	int dp=0;
	int srcpitch=src->width*4;
	int destpitch=width*4;
	s_b=s_g=s_r=s_a=0;
	for (int j=0;j<dh;j++) {
		ox=x;oy=y;
		for (int i=0;i<dw;i++) {
			if (dlx+dcx+i<0 || dlx+dcx+i>=width || dly+dcy+j<0 || dly+dcy+j>=height) {
				goto _next;
			}
			int xa[4],ya[4],pa[4][4];
			if (x>0) {
				xa[0]=int(x+0.0000001)-1;
				xa[1]=int(x+0.0000001);
				xa[2]=int(x+0.9999999);
				xa[3]=int(x+0.9999999)+1;
			} else {
				xa[0]=int(x-0.0000001)-1;
				xa[1]=int(x-0.0000001);
				xa[2]=int(x-0.9999999);
				xa[3]=int(x-0.9999999)+1;
			}
			if (y>0) {
				ya[0]=int(y+0.0000001)-1;
				ya[1]=int(y+0.0000001);
				ya[2]=int(y+0.9999999);
				ya[3]=int(y+0.9999999)+1;
			} else {
				ya[0]=int(y-0.0000001)-1;
				ya[1]=int(y-0.0000001);
				ya[2]=int(y-0.9999999);
				ya[3]=int(y-0.9999999)+1;
			}
			//ߖTPU_m
			bool fl=false;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					if (xa[ii]+sw/2<0 || xa[ii]+sw/2>=sw || ya[jj]+sh/2<0 || ya[jj]+sh/2>=sh) {
						pa[ii][jj]=-1;
					} else {
						fl=true;
						pa[ii][jj]=(xa[ii]+sw/2)*4+(ya[jj]+sh/2)*srcpitch;
					}
				}
			}
			if (!fl) goto _next;
			double w,p0,p1,p2,p3,d[4][4];
			p0=p1=p2=p3=0.0;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					double dx=double(xa[ii])-x;
					double dy=double(ya[jj])-y;
					d[ii][jj]=sqrt(dx*dx+dy*dy);
				}
			}
			double tw=0.0;
			double r;
			for (int jj=0;jj<4;jj++) {
				for (int ii=0;ii<4;ii++) {
					double dd=d[ii][jj];
					r=3.141592653589793238463*dd;
					w=(sin(r)/r)*2.0*sin(r/2.0)/r;
					tw+=w;
					if (pa[ii][jj]==-1) continue;
					p0+=srcb[pa[ii][jj]]*w;
					p1+=srcb[pa[ii][jj]+1]*w;
					p2+=srcb[pa[ii][jj]+2]*w;
					p3+=srcb[pa[ii][jj]+3]*w;
				}
			}
			p0/=tw;
			p1/=tw;
			p2/=tw;
			p3/=tw;
			int p=(int)p3;
			if (p<0) p=0; else if (p>255) p=255;
			s_a=srcalpha*p/255;
			s_b=(int)p0;
			if (s_b<0) s_b=0; else if (s_b>255) s_b=255;
			s_g=(int)p1;
			if (s_g<0) s_g=0; else if (s_g>255) s_g=255;
			s_r=(int)p2;
			if (s_r<0) s_r=0; else if (s_r>255) s_r=255;
			int c;
			switch (md) {
				case BM_ALPHA:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*s_a/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*s_a/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*s_a/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
				case BM_ADD:
					c=destb[ptr+dp]+s_b;
					destb[ptr+dp]=(c>255)?255:c;
					c=destb[ptr+dp+1]+s_g;
					destb[ptr+dp+1]=(c>255)?255:c;
					c=destb[ptr+dp+2]+s_r;
					destb[ptr+dp+2]=(c>255)?255:c;
					destb[ptr+dp+3]=255;
					break;
				case BM_COPY:
					destb[ptr+dp]=(s_b-destb[ptr+dp])*srcalpha/255+destb[ptr+dp];
					destb[ptr+dp+1]=(s_g-destb[ptr+dp+1])*srcalpha/255+destb[ptr+dp+1];
					destb[ptr+dp+2]=(s_r-destb[ptr+dp+2])*srcalpha/255+destb[ptr+dp+2];
					destb[ptr+dp+3]=255;
					break;
			}
_next:
			dp+=4;
			x+=m._00;y+=m._10;
		}
		x=ox+m._01;y=oy+m._11;
		dp=0;
		ptr+=destpitch;
	}
}
